from __future__ import division, print_function, absolute_import, unicode_literals

import functools
import time
from builtins import bytes, str, range, str
from bisect import bisect_left
import os
import numpy
import tables
from .models import KeyWrapper
import logging

logger = logging.getLogger(__name__)


class SuffixIndexBuilderStringCol(object):
    MEM_PER_CHUNK = 50 * 2**20

    def __init__(self, tab, col, index_group, ignore_case):
        if not isinstance(tab, tables.Table):
            raise ValueError("tab argument must be a tables.Table object")
        if col not in tab.colnames:
            raise ValueError("column '{}' does not exist in {}".format(col, tab))
        self.h5 = tab._v_file
        self.tab = tab
        self.col = col
        self.index_group = index_group if isinstance(index_group, tables.Group) else self.h5.get_node(index_group)
        self.ignore_case = ignore_case

    def check_column_types_or_raise(self):
        if not numpy.issubdtype(self.tab.coldtypes[self.col].type, numpy.bytes_):
            raise TypeError("column '{}' must be a character type column".format(self.col))

    def get_expected_index_length(self):
        return self.tab.coldtypes[self.col].itemsize * len(self.tab)

    def _arrayname(self, kind):
        return self.col + "_" + kind

    def _remove_aux_arrays(self):
        for kind in ("suffix", "buffer", "offset"):
            try:
                n = self.get_aux_array_handle(kind)
                index_group_path = self.index_group._v_pathname
                if os.path.commonprefix([index_group_path, n._v_pathname]) == index_group_path:
                    # auxilary buffer hangs on the index_group. we can remove it
                    logger.info("removing existing auxillary node: {}".format(n._v_pathname))
                    n.remove()
                    try:
                        self.h5.del_node_attr(self.index_group, self._arrayname(kind))
                    except AttributeError:
                        pass
            except tables.NoSuchNodeError:
                pass

    def create_aux_arrays(self):
        for kind, typ in (
            ("buffer", tables.StringAtom(1)),
            ("offset", tables.UInt32Atom()),
        ):
            arr_name = self._arrayname(kind)
            exp_rows = self.get_expected_index_length() if kind == "buffer" else len(self.tab)
            arr = self.h5.create_earray(self.index_group, arr_name, typ, (0,), expectedrows=exp_rows)
            self.h5.set_node_attr(self.index_group, arr_name, arr._v_pathname)

    def get_aux_array_handle(self, kind):
        if kind not in ("buffer", "offset", "suffix"):
            raise ValueError("Not a valid handle for suffix index")
        attr = self._arrayname(kind)
        try:
            path = self.h5.get_node_attr(self.index_group, attr)
        except AttributeError:
            raise tables.NoSuchNodeError("'{}' does not exist".format(attr))
        return self.h5.get_node(path)

    def build_index_buffer(self):
        chunksize = int(self.MEM_PER_CHUNK / self.tab.coldtypes[self.col].itemsize)
        total_offset = 0
        buffer, offset = (self.get_aux_array_handle(kind) for kind in ("buffer", "offset"))
        for chunk_start in range(0, len(self.tab), chunksize):
            # load fixed width string col in numpy array, compute actual string lengths and build
            # a \x00 delimited buffer of all values
            col_data = self.tab.read(start=chunk_start, stop=chunk_start + chunksize, field=self.col)
            data_lens = numpy.char.str_len(col_data)
            data_as_long_arr = col_data.view("S1")
            tot_len = (data_lens + 1).sum()
            buf = numpy.zeros(tot_len, dtype="S1")
            t = (data_lens + 1).cumsum()
            starts = numpy.roll(t, 1)
            starts[0] = 0
            ends = t - 1
            stride = col_data.strides[0]
            for i in range(len(col_data)):
                buf[starts[i] : ends[i]] = data_as_long_arr[(stride * i) : (stride * i + data_lens[i])]
            if self.ignore_case:
                buf = numpy.char.lower(buf)
            # update global offset and append chunked buffer
            starts += total_offset
            buffer.append(buf)
            offset.append(starts)
            total_offset += tot_len

    def build_suffix_array(self):
        from PySAIS import sais

        data = self.get_aux_array_handle("buffer")[:]
        suffix = sais(data) if len(data) > 0 else data
        arr = self.h5.create_earray(self.index_group, self._arrayname("suffix"), obj=suffix)
        self.h5.set_node_attr(self.index_group, self._arrayname("suffix"), arr._v_pathname)
        self.h5.set_node_attr(self.index_group, self.col + "_ignore_case", self.ignore_case)
        self.h5.set_node_attr(self.tab, self.col + "_suffixindexnode", self.index_group._v_pathname)

    def __call__(self, force=True):
        self.check_column_types_or_raise()
        if force:
            self._remove_aux_arrays()
        self.create_aux_arrays()
        self.build_index_buffer()
        self.build_suffix_array()


class SuffixIndexBuilderVarStringCol(SuffixIndexBuilderStringCol):
    def __init__(self, tab, col, buffer, index_group, ignore_case):
        if not isinstance(buffer, tables.CArray):
            raise TypeError("buffer argument must be a tables.CArray instance")
        super(SuffixIndexBuilderVarStringCol, self).__init__(tab, col, index_group, ignore_case)
        self.orig_buffer = buffer

    def check_column_types_or_raise(self):
        if not numpy.issubdtype(self.orig_buffer.dtype.type, numpy.bytes_):
            raise TypeError("buffer '{}' must be a character type column".format(self.orig_buffer._v_pathname))
        if not numpy.issubdtype(self.tab.coldtypes[self.col].type, numpy.integer):
            raise TypeError(
                "column '{}' must be an integer argument with offsets into the '{}' buffer array".format(
                    self.col, self.orig_buffer._v_pathname
                )
            )

    def get_expected_index_length(self):
        return len(self.orig_buffer) + len(self.tab)  # extra char per row for delimiting

    def create_aux_arrays(self):
        for kind, typ in (
            ("buffer", tables.StringAtom(1)),
            ("offset", tables.UInt32Atom()),
        ):
            arr_name = self._arrayname(kind)
            if not self.ignore_case and kind == "buffer":
                self.h5.set_node_attr(self.index_group, arr_name, self.orig_buffer._v_pathname)
            else:
                exp_rows = self.get_expected_index_length() if kind == "buffer" else len(self.tab)
                arr = self.h5.create_earray(self.index_group, arr_name, typ, (0,), expectedrows=exp_rows)
                self.h5.set_node_attr(self.index_group, arr_name, arr._v_pathname)

    def build_index_buffer(self):
        off_data = self.h5.get_node(self.tab).read(field=self.col)
        # first, we need to fix cases where off_data is 0 in the middle of an array,
        # simply because it has not been set as its length is also 0. We do this
        # by setting the offsets to the next starting offset of a non-zero length
        # entry (by copying the next value from the back)
        # set every zero position except [0] to the one following in the buffer,
        # in case the buffer ends with an uninitialized value, set it to the
        # length of the entire buffer.
        if off_data[-1] == 0:
            off_data[-1] = len(self.orig_buffer)
            # if now off_data[-1] is still 0, then there is actually no buffer data at all
            if off_data[-1] == 0:
                off_arr = self.get_aux_array_handle("offset")
                off_arr.append(off_data)
                return
        zero_positions = numpy.where(off_data == 0)[0]
        for idx in zero_positions[:0:-1]:
            off_data[idx] = off_data[idx + 1]

        if self.ignore_case:
            starts = off_data
            # now, set the stop position by using the next start postion (shift array by 1)
            stops = numpy.roll(off_data, -1)
            stops[-1] = len(self.orig_buffer)
            buf_arr = self.get_aux_array_handle("buffer")
            CHUNKSIZE = 2**22  # 4MB
            tmp_buf = numpy.zeros(CHUNKSIZE, dtype="S1")
            tmp_buf_idx = 0
            for i in range(len(off_data)):
                if stops[i] - starts[i] + tmp_buf_idx >= CHUNKSIZE:
                    buf_arr.append(numpy.char.lower(tmp_buf[0:tmp_buf_idx]))
                    tmp_buf = numpy.zeros(CHUNKSIZE, dtype="S1")
                    tmp_buf_idx = 0
                tmp_buf[tmp_buf_idx : (tmp_buf_idx + stops[i] - starts[i])] = self.orig_buffer[starts[i] : stops[i]]
                tmp_buf_idx += stops[i] - starts[i] + 1  # keep one extra '\x00' as separator
            if tmp_buf_idx > 0:
                buf_arr.append(numpy.char.lower(tmp_buf[0:tmp_buf_idx]))
            # add one extra position to each string that contains a '\x00'
            off_data += numpy.arange(0, len(off_data), 1, dtype=off_data.dtype)

        off_arr = self.get_aux_array_handle("offset")
        off_arr.append(off_data)


def create_or_load_index_group(tab, index_group=None):
    f = tab._v_file
    if index_group is None:
        idx_prefix, idx_name = tab._v_parent, "_si_" + tab._v_name
    else:
        idx_prefix, idx_name = index_group.rsplit("/", 1)

    try:
        return f.get_node(idx_prefix, name=idx_name)
    except tables.NoSuchNodeError:
        return f.create_group(idx_prefix, idx_name)


def create_suffix_index(tab, col, buffer=None, index_group=None, ignore_case=True, force=False):
    """Create a suffix array from the given table column (or buffer).

    This function creates for a given fix-width char column a suffix array for
    fast lookup of (partial) matches. In this case, the `buffer` argument
    should be None. Alternatively, if a index should be build from a variable
    length column, i.e. a buffer array with offset and length attributes in
    the table, one should pass the offset column as `col` argument and the
    char buffer as `buffer` argument.

    By default, the class will create an array named <colname>_suffix,
    <colname>_buffer and <colname>_offset in a group node that is named and
    located either according to the argument `index_group` or as a sibling
    to `tab` named <tab>_Indexes.

    `ignore_case` will indicate that the index should be build case insensitive.
    In case an index is build from a variable length column (see above what
    we understand this is in this context) and `ignore_case` is False, the
    buffer and offset array will not be built again.

    :param tab: a handle to a h5 table node
    :param str col: the name the column to be indexed
    :param buffer: handle to buffer array
    :param str index_group: path to existing or non-existing group node
        where index data will be stored.
    :param bool ignore_case: case insensitive lookup
    :param force: overwrite existing suffix array data"""
    if not isinstance(tab, tables.Table):
        raise TypeError("tab arguments must be a tables.Table (is a {})".format(type(tab)))
    if not isinstance(col, str) or col not in tab.colnames:
        raise ValueError("table {} does not have a column named '{}'".format(tab, col))
    typ = tab.coldtypes[col]
    if numpy.issubdtype(typ.type, numpy.bytes_):
        if typ.itemsize < 3:
            raise TypeError("suffix arrays should be created for longer string columns only")
        suffixbuilder = SuffixIndexBuilderStringCol(
            tab,
            col,
            create_or_load_index_group(tab, index_group),
            ignore_case=ignore_case,
        )
    elif numpy.issubdtype(typ.type, numpy.integer):
        if buffer is None:
            raise ValueError("buffer array must be specified for numeric table columns")
        if isinstance(buffer, str):
            buffer = tab._v_file.get_node(buffer)
        if not (
            isinstance(buffer, tables.CArray) and numpy.issubdtype(buffer, numpy.bytes_) and buffer.dtype.itemsize == 1
        ):
            raise TypeError("buffer array must by a character tables.CArray instance (itemsize == 1)")
        suffixbuilder = SuffixIndexBuilderVarStringCol(
            tab,
            col,
            buffer,
            create_or_load_index_group(tab, index_group),
            ignore_case=ignore_case,
        )
    else:
        raise TypeError("unsupported type of column to index")
    return suffixbuilder()


class SuffixSearcher(object):
    @classmethod
    def from_tablecolumn(cls, table, column, ignore_case=False):
        h5 = table._v_file
        try:
            idx_node = h5.get_node_attr(table, column + "_suffixindexnode")
            idx_node = h5.get_node(idx_node)
        except (AttributeError, tables.NoSuchNodeError) as e:
            raise SuffixIndexError("Column {} of table {} does not seem to have a suffix index".format(column, table))
        try:
            suffix_arr = h5.get_node(idx_node, column + "_suffix")
            buffer_arr = h5.get_node(h5.get_node_attr(idx_node, column + "_buffer"))
            offset_arr = h5.get_node(h5.get_node_attr(idx_node, column + "_offset"))
        except (tables.NoSuchNodeError, AttributeError) as e:
            raise SuffixIndexInconsitency("not all suffix index elements available: {}".format(e))
        try:
            ignore_case = bool(h5.get_node_attr(idx_node, column + "_ignore_case"))
        except AttributeError:
            ignore_case = False
        return cls(suffix_arr, buffer_arr, offset_arr, ignore_case)

    @classmethod
    def from_index_node(cls, index_node, buffer=None, offset=None):
        import warnings

        warnings.warn(
            "initializing SuffixSearcher this way is deprecated. Use from_tablecolumn(table, columnname) instead"
        )
        if not isinstance(index_node, tables.Group):
            raise TypeError("expected a tables.Group node pointing to the index")
        try:
            buffer_arr = buffer if buffer else index_node._f_get_child("buffer")
            suffix_arr = index_node._f_get_child("suffix")
            offset_arr = offset if offset else index_node._f_get_child("offset")
            return cls(suffix_arr, buffer_arr, offset_arr, ignore_case=True)
        except tables.NoSuchNodeError as e:
            raise SuffixIndexInconsitency("suffix array data missing: {}".format(e))

    def __init__(self, suffix, buffer, offset, ignore_case):
        self.suffix_arr = suffix
        self.buffer_arr = buffer
        self.offset_arr = offset[:]
        self.ignore_case = bool(ignore_case)

    @functools.lru_cache(maxsize=16)
    def _get_index_range(self, query):
        if isinstance(query, str):
            query = query.encode("utf-8")
        if self.ignore_case:
            query = query.lower()
        n = len(query)
        if n > 0:
            slicer = KeyWrapper(self.suffix_arr, key=lambda i: self.buffer_arr[i : (i + n)].tobytes())
            ii = bisect_left(slicer, query)
            if ii and ii < len(self.suffix_arr) and (slicer[ii] == query):
                query_after = query[:-1] + chr(query[-1] + 1).encode("utf-8")
                jj = bisect_left(slicer, query_after)
                if (jj < len(self.suffix_arr) and slicer[jj] == query) or (slicer[jj - 1] != query):
                    raise RuntimeError("index broken, should not happen")
                return slice(ii, jj)
        return slice(0, 0)

    def count(self, query):
        index_range = self._get_index_range(query)
        return index_range.stop - index_range.start

    def find(self, query, limit=None):
        index_range = self._get_index_range(query)
        if index_range.stop - index_range.start == 0:
            return numpy.array([], self.offset_arr.dtype)

        if limit is not None:
            index_range = slice(index_range.start, min(index_range.stop, index_range.start + limit))

        # Find row numbers
        res = numpy.searchsorted(self.offset_arr, self.suffix_arr[index_range] + 1) - 1
        return res


class SuffixIndexError(Exception):
    pass


class SuffixIndexInconsitency(SuffixIndexError):
    pass



import collections
from collections import Counter, defaultdict

import numpy
import pandas
import pathlib
from property_manager import lazy_property
from ..browser.db import Database
from ..browser.geneontology import GOAspect
from scipy.stats import fisher_exact
from statsmodels.stats.multitest import multipletests
from sklearn.manifold import MDS
from typing import Iterable, Union
import plotly
import plotly.graph_objects as go
import itertools
import pandas as pd
import logging

logger = logging.getLogger(__name__)
LONGONT = {"F": "MF", "P": "BP", "C": "CC"}


class GOEnrichmentAnalysis(object):
    def __init__(self, db, population, annots, most_specific=False, ensure_term=False):
        self._db = db
        self._go = db.gene_ontology

        self._annots = self._annot_coverage(annots, most_specific, ensure_term)
        self._with_annot = set(self._annots.keys())
        self._pop = set(population) & self._with_annot
        self._pop_counts = self._get_term_counts(self._pop)

    def _annot_coverage(self, annots, most_specific, ensure_term):
        if not most_specific or ensure_term:
            #  first get the most specific set for each entry
            for g in annots:
                if ensure_term:
                    annots[g] = set(map(self._go.term_by_id, annots[g]))

                z = set()
                for t in annots[g]:
                    if (len(annots[g] & self._go.get_subterms(t))) == 1:
                        z.add(t)
                annots[g] = z

        #  create vocabulary for goea
        vocab = set(itertools.chain.from_iterable(annots.values()))
        # now expand into the vocab for each term
        for g in annots:
            z = set()
            for t in annots[g]:
                z.update(self._go.get_superterms_incl_queryterm(t) & vocab)
            annots[g] = z

        return annots

    @lazy_property
    def _term_to_entry(self):
        z = defaultdict(set)
        for k, v in self._annots.items():
            for t in v:
                z[t].add(k)
        return z

    def _get_term_counts(self, ents):
        # counts a set of terms
        z = Counter()
        for t, t_ents in self._term_to_entry.items():
            n = len(t_ents & ents)
            if n > 0:
                z[t] = n
        return z

    def run_study(self, study, alpha=0.05, correction=None):
        """
        Run a GOEA with the study set as a foreground, with the already loaded background.
        """

        def do(study, study_counts):
            study_n = len(study)
            pop_n = len(self._pop)

            for t in study_counts:
                study_count = study_counts[t]
                pop_count = self._pop_counts[t]

                study_ratio = study_count / study_n
                pop_ratio = pop_count / pop_n

                # do we need to be able to compute the study_entries?
                study_entries = self._term_to_entry[t] & study

                if study_ratio >= pop_ratio:
                    # only test enrichment
                    a = study_count
                    b = study_n - study_count
                    c = pop_count - study_count
                    d = pop_n - pop_count - b

                    p = fisher_exact([[a, b], [c, d]], alternative="greater")[1]
                    yield (t, study_count, pop_count, study_entries, study_n, pop_n, p)

        if correction is None:
            correction = "fdr_bh"
        if correction not in ("fdr_bh", "bonferroni"):
            raise ValueError(f"correction parameter must be 'fdr_bh' or 'bonferroni': {correction}")
        study = set(study) & self._with_annot
        study_counts = self._get_term_counts(study)

        df = pd.DataFrame(
            do(study, study_counts),
            columns=[
                "GO_ID",
                "study_count",
                "population_count",
                "study_entries_with_go_term",
                "study_size",
                "pop_n",
                "p_uncorrected",
            ],
        )
        if len(df) == 0:
            return None

        # corrected p values
        df["p_bonferroni"] = multipletests(df.p_uncorrected.values, alpha=alpha, method="bonferroni")[1]
        df["p_fdr_bh"] = multipletests(df.p_uncorrected.values, alpha=alpha, method="fdr_bh")[1]
        if correction == "fdr_bh":
            df = df[df["p_fdr_bh"] <= alpha]
        elif correction == "bonferroni":
            df = df[df["p_bonferroni"] <= alpha]

        # results
        df["GO_name"] = df["GO_ID"].apply(lambda t: t.name)
        df["GO_depth"] = df["GO_ID"].apply(lambda t: t.min_depth)
        df["GO_aspect"] = df["GO_ID"].apply(lambda t: LONGONT[GOAspect.to_char(t.aspect)])
        df["GO_ID"] = df["GO_ID"].apply(lambda t: str(t))

        df["study_proportion"] = df.apply(lambda x: (x["study_count"] / x["study_size"]), axis=1)
        df["study_ratio"] = df.apply(lambda x: "{} / {}".format(x["study_count"], x["study_size"]), axis=1)
        df["study_entries_with_go_term"] = df["study_entries_with_go_term"].apply(
            lambda x: ", ".join(sorted(map(lambda e: e.decode() if isinstance(e, bytes) else str(e), x)))
        )

        df["population_proportion"] = df.apply(lambda x: (x["population_count"] / x["pop_n"]), axis=1)
        df["population_ratio"] = df.apply(lambda x: "{} / {}".format(x["population_count"], x["pop_n"]), axis=1)
        df["population_size"] = df["pop_n"]
        df["fold_change"] = df["study_proportion"] / df["population_proportion"]

        cols = [
            "GO_ID",
            "GO_name",
            "GO_aspect",
            "GO_depth",
            "p_uncorrected",
            "p_bonferroni",
            "p_fdr_bh",
            "study_count",
            "study_size",
            "study_ratio",
            "study_proportion",
            "population_count",
            "population_size",
            "population_ratio",
            "population_proportion",
            "fold_change",
            "study_entries_with_go_term",
        ]
        df = df[cols]

        return df.sort_values(
            [f"p_{correction}", "fold_change", "p_uncorrected"],
            ascending=[True, False, True],
        ).reset_index(drop=True)


def extant_species_go_enrichment(db: Database, foreground_entries: Iterable[int], alpha=0.05) -> pd.DataFrame:
    foreground = sorted(set(foreground_entries))
    omaid_mapper = db.id_mapper["OMA"]
    rng = omaid_mapper.genome_range(foreground[0])
    if rng[1] < foreground[-1]:
        fnd_species = collections.Counter(map(lambda enr: omaid_mapper.map_entry_nr(enr)[:5], foreground_entries))
        raise ValueError(
            f"foreground_entries must all be from the same species. "
            f"Found {len(fnd_species)} different ones (top 3: {fnd_species.most_common(3)}"
        )
    anno = defaultdict(set)
    for row in db.get_gene_ontology_annotations(rng[0], stop=rng[1] + 1):
        anno[omaid_mapper.map_entry_nr(int(row["EntryNr"]))].add(int(row["TermNr"]))
    main_iso = [omaid_mapper.map_entry_nr(int(row["EntryNr"])) for row in db._main_isoform_iter(rng[0], rng[1])]
    goea = GOEnrichmentAnalysis(db, main_iso, annots=anno, ensure_term=True)
    logger.debug("go_traverse_cache: %s", goea._go._traverseGraph.cache_info())
    return goea.run_study(map(omaid_mapper.map_entry_nr, foreground), alpha=alpha)


def ancestral_species_go_enrichment(
    db: Database,
    level: Union[str, bytes, int],
    foreground_hogs: Iterable[Union[str, bytes]],
    alpha=0.05,
    score_cutoff=0.1,
) -> pd.DataFrame:
    anc_node = db._ancestral_node(level)
    hog_ids = anc_node.Hogs.read(field="ID")
    foreground_hogs = numpy.fromiter(
        set(x.encode("utf-8") if isinstance(x, str) else x for x in foreground_hogs),
        dtype="S255",
    )
    print(foreground_hogs)
    if not numpy.isin(foreground_hogs, hog_ids).all():
        logger.warning(f"not all hogs found at the requested level: {numpy.isin(foreground_hogs, hog_ids)}")
        print(numpy.isin(foreground_hogs, hog_ids))
        raise ValueError("foreground_hogs must all be from the same level")
    annots = defaultdict(set)
    for row in anc_node.GeneOntology:
        if row["RawScore"] >= score_cutoff:
            annots[hog_ids[row["HogRow"]]].add(int(row["TermNr"]))
    goea = GOEnrichmentAnalysis(db, hog_ids, annots=annots, ensure_term=True)
    return goea.run_study(foreground_hogs, alpha=alpha)


class GO_MDS:
    def __init__(self, db):
        self.db = db

    def __call__(self, terms):
        return self.mds(terms)

    def mds(self, terms):
        """
        Perform MDS on the semantic similarity between the terms
        """
        # split the terms into different aspects first
        terms_by_aspect = defaultdict(set)
        for t in map(self.db.freq_aware_gene_ontology.term_by_id, terms):
            terms_by_aspect[t.aspect].add(t)

        # for each aspect, perform the
        mds = {}
        for aspect, subterms in terms_by_aspect.items():
            aspect_desc = GOAspect.to_string(aspect)

            subterms = sorted(map(lambda t: t.id, subterms))

            n = len(subterms)
            ss = numpy.ones((n, n), dtype=numpy.float64)
            for i in range(n):
                for j in range(i + 1, n):
                    ss[i, j] = ss[j, i] = self.db.freq_aware_gene_ontology.semantic_similarity(subterms[i], subterms[j])

            #  now perform a 2d MDS
            xy = MDS(dissimilarity="precomputed", normalized_stress="auto").fit(10 * (1 - ss)).embedding_
            mds[aspect_desc] = {t: tuple(xy[i]) for (i, t) in enumerate(subterms)}
        return mds


def generate_plots(enr_df, db, folder=None, pval_col="p_bonferroni"):
    """generate plots MDS of similarities of GO terms with significance and information content.

    The dataframes with the resulting MDS embedings is also returned."""
    if folder is None:
        folder = ""
    folder = pathlib.Path(folder)
    mds = GO_MDS(db)
    xy = mds.mds(enr_df["GO_ID"])
    per_aspect = {}
    for aspect, terms in xy.items():
        labels = sorted(terms.keys())
        df = pandas.DataFrame(
            {
                "id": labels,
                "MDS1": [terms[x][0] for x in labels],
                "MDS2": [terms[x][1] for x in labels],
            }
        )
        df["GO_term"] = df["id"].apply(db.freq_aware_gene_ontology.ensure_term)
        df["id"] = df["GO_term"].apply(str)
        df["name"] = df["GO_term"].apply(lambda t: t.name)
        df["ic"] = df["GO_term"].apply(lambda t: db.freq_aware_gene_ontology.ic(t))
        df = pandas.merge(df, enr_df[["GO_ID", pval_col]], left_on="id", right_on="GO_ID")
        df.drop(columns=["id", "GO_term"], inplace=True)

        fig = go.Figure()
        fig.add_trace(
            go.Scatter(
                x=df["MDS1"],
                y=df["MDS2"],
                mode="markers",
                marker={
                    "size": 200 / (df["ic"] + 3),
                    "color": numpy.log10(df[pval_col]),
                    "colorbar": {"title": f"p-value ({pval_col}) (log10 scale)"},
                },
                opacity=0.9,
                customdata=numpy.stack((df["GO_ID"], df["name"], df["ic"], df[pval_col]), axis=-1),
                hovertemplate=(
                    "<b>ID</b>: %{customdata[0]}<br>"
                    + "<b>Name</b>: %{customdata[1]}<br>"
                    + "<b>Information Content</b>: %{customdata[2]:,.3f}<br>"
                    + "<b>p-value</b>: %{customdata[3]:,.3e}<br>"
                    + "<extra></extra>"
                ),
            )
        )
        fig.update_layout(
            xaxis={"title": "Semantic Space X"},
            yaxis={"title": "Semantic Space Y"},
            title=f"{aspect.title()}",
        )
        plotly.offline.plot(fig, filename=str(folder / f"{aspect}.html"))
        per_aspect[aspect] = df
    return per_aspect



import pandas
import logging
import collections
import numpy as np
from skfuzzy import control as ctrl
from skfuzzy import gaussmf
import sklearn
import sklearn.preprocessing
import tables

try:
    from tqdm import tqdm
except ImportError:
    tqdm = lambda x, **kwargs: x
logger = logging.getLogger(__name__)


def define_universe(df):
    # New Antecedent/Consequent objects hold universe variables and membership functions
    distance = ctrl.Antecedent(np.arange(0, df["Distance"].max() + 0.01, 0.01), "distance")
    synteny = ctrl.Antecedent(np.arange(0, 1.01, 0.01), "synteny_score")
    total_nb_hom = ctrl.Antecedent(np.arange(2, df["TotalCopyNr"].max() + 1, 1), "total_nb_homoeologs")
    conf = ctrl.Consequent(np.arange(0, 101, 1), "conf")
    return distance, synteny, total_nb_hom, conf


def create_fuzzy_rules(distance, synteny, total_nb_hom, conf):
    """Takes the antecedent and consequent objects as input"""

    # very low confidence
    rule1 = ctrl.Rule(synteny["low"] & distance["high"] & total_nb_hom["high"], conf["very_low"])

    # low confidence
    rule2 = ctrl.Rule(
        (synteny["low"] & distance["high"] & total_nb_hom["low"])
        | (synteny["low"] & distance["high"] & total_nb_hom["med"])
        | (synteny["low"] & distance["med"] & total_nb_hom["high"])
        | (synteny["med"] & distance["high"] & total_nb_hom["high"]),
        conf["low"],
    )

    # medium confidence
    rule3 = ctrl.Rule(
        (synteny["high"] & distance["high"] & total_nb_hom["high"])
        | (synteny["low"] & distance["med"] & total_nb_hom["low"])
        | (synteny["low"] & distance["med"] & total_nb_hom["med"])
        | (synteny["med"] & distance["high"] & total_nb_hom["low"])
        | (synteny["med"] & distance["high"] & total_nb_hom["med"])
        | (synteny["med"] & distance["med"] & total_nb_hom["high"])
        | (synteny["med"] & distance["med"] & total_nb_hom["low"])
        | (synteny["med"] & distance["med"] & total_nb_hom["med"])
        | (synteny["low"] & distance["low"] & total_nb_hom["low"])
        | (synteny["low"] & distance["low"] & total_nb_hom["med"])
        | (synteny["low"] & distance["low"] & total_nb_hom["high"]),
        conf["med"],
    )

    # high confidence
    rule4 = ctrl.Rule(
        (synteny["high"] & distance["high"] & total_nb_hom["low"])
        | (synteny["high"] & distance["high"] & total_nb_hom["med"])
        | (synteny["high"] & distance["low"] & total_nb_hom["high"])
        | (synteny["high"] & distance["low"] & total_nb_hom["med"])
        | (synteny["high"] & distance["med"] & total_nb_hom["high"])
        | (synteny["high"] & distance["med"] & total_nb_hom["low"])
        | (synteny["high"] & distance["med"] & total_nb_hom["med"])
        | (synteny["med"] & distance["low"] & total_nb_hom["high"])
        | (synteny["med"] & distance["low"] & total_nb_hom["med"])
        | (synteny["med"] & distance["low"] & total_nb_hom["low"]),
        conf["high"],
    )

    # very high confidence
    rule5 = ctrl.Rule(synteny["high"] & distance["low"] & total_nb_hom["low"], conf["very_high"])
    return [rule1, rule2, rule3, rule4, rule5]


def get_distance_mf(df, distance):
    # here, the first numnber is the universe, second number the central point, third the standard deviation
    distance["low"] = gaussmf(distance.universe, 0, (df["Distance"].max() / 10))

    distance["med"] = gaussmf(distance.universe, (df["Distance"].max() / 4), (df["Distance"].max() / 10))

    distance["high"] = gaussmf(distance.universe, df["Distance"].max(), (df["Distance"].max() / 2.5))
    return distance


def get_synteny_mf(df, synteny, view=False):
    # synteny (gaussian)
    synteny["low"] = gaussmf(synteny.universe, 0, 0.15)
    synteny["med"] = gaussmf(synteny.universe, 0.3, 0.15)
    synteny["high"] = gaussmf(synteny.universe, 1, 0.4)
    return synteny


def get_total_nb_hom_mf(df, total_nb_hom):
    copy_nr_median = df["TotalCopyNr"].median()
    total_nb_hom["low"] = gaussmf(total_nb_hom.universe, copy_nr_median, copy_nr_median)

    total_nb_hom["med"] = gaussmf(total_nb_hom.universe, 4 * copy_nr_median, 1.5 * copy_nr_median)

    total_nb_hom["high"] = gaussmf(total_nb_hom.universe, df["TotalCopyNr"].max(), df["TotalCopyNr"].max() / 2.5)
    return total_nb_hom


def get_conf_mf(df, conf):
    # confidence (gaussian)
    conf["very_low"] = gaussmf(conf.universe, 0, 20)
    conf["low"] = gaussmf(conf.universe, 50, 10)
    conf["med"] = gaussmf(conf.universe, 70, 10)
    conf["high"] = gaussmf(conf.universe, 90, 10)
    conf["very_high"] = gaussmf(conf.universe, 100, 10)
    return conf


def get_conf_score(simulation, input_dic):
    """This function takes the simulation and outputs confidence score
    'input_dic' is a dictionary of the inputs for a homoeolog pair"""

    simulation.inputs(input_dic)
    simulation.compute()
    return simulation.output["conf"]


class HomeologsConfidenceCalculator(object):
    fields_to_load = ["EntryNr1", "EntryNr2", "SyntenyConservationLocal", "Distance"]

    def __init__(self, h5_handle, genome):
        self.h5_handle = h5_handle
        self.genome = genome
        if isinstance(h5_handle, tables.File):
            self.h5_handle = h5_handle
        elif isinstance(h5_handle, (str, bytes)):
            self.h5_handle = tables.open_file(h5_handle, "r")
        else:
            raise TypeError("expected h5_handle to be either h5-file handle or a path to file")

        genome_row = next(self.h5_handle.root.Genome.where("UniProtSpeciesCode == genome"))
        self.genome_range = (
            int(genome_row["EntryOff"]) + 1,
            int(genome_row["EntryOff"] + genome_row["TotEntries"]),
        )
        genome_df = pandas.DataFrame(
            self.h5_handle.root.Protein.Entries.read_where(
                "(EntryNr >= {}) & (EntryNr <= {})".format(*self.genome_range)
            )
        )
        self.genome_df = genome_df[
            (genome_df["AltSpliceVariant"] == 0) | (genome_df["AltSpliceVariant"] == genome_df["EntryNr"])
        ]
        self.genome_df.reset_index(inplace=True)
        self.relations_df = self._load_pairwise_relations()

    def _load_pairwise_relations(self):
        """load the homoeologous relations of the cannonical splice variants only
        The method returns a pandas dataframe with the relations."""
        df = pandas.DataFrame(
            self.h5_handle.get_node("/PairwiseRelation/{}/within".format(self.genome)).read_where("RelType == 5")
        )
        df = df[df["EntryNr1"].isin(self.genome_df["EntryNr"]) & df["EntryNr2"].isin(self.genome_df["EntryNr"])]
        return df[self.fields_to_load]

    def _count_homeologs_per_entry(self, df):
        return collections.Counter(df["EntryNr1"])

    def _augment_dataframe_with_all_features(self, df):
        counts = self._count_homeologs_per_entry(df)
        df["TotalCopyNr"] = df.apply(lambda x: counts[x["EntryNr1"]] + counts[x["EntryNr2"]], axis=1)
        df.loc[df.SyntenyConservationLocal < 0, "SyntenyConservationLocal"] = 0
        return df

    def calculate_scores(self):
        # load dataframe
        df = self.relations_df
        df = self._augment_dataframe_with_all_features(df)

        distanceObj, syntenyObj, total_nb_homObj, confObj = define_universe(df)
        distance = get_distance_mf(df, distanceObj)
        synteny = get_synteny_mf(df, syntenyObj)
        total_nb_hom = get_total_nb_hom_mf(df, total_nb_homObj)
        conf = get_conf_mf(df, confObj)

        # create simulation
        rules = create_fuzzy_rules(distance, synteny, total_nb_hom, conf)
        control_system = ctrl.ControlSystem(rules)
        simulation = ctrl.ControlSystemSimulation(control_system)

        def defuzzify(row):
            return get_conf_score(
                simulation,
                {
                    "distance": row["Distance"],
                    "synteny_score": row["SyntenyConservationLocal"],
                    "total_nb_homoeologs": row["TotalCopyNr"],
                },
            )

        df["fuzzy_confidence"] = df.apply(defuzzify, axis=1)

        # scale the confidence between minimum value and 100
        min_max_scaler = sklearn.preprocessing.MinMaxScaler(feature_range=(df["fuzzy_confidence"].min(), 100))
        df["fuzzy_confidence_scaled"] = min_max_scaler.fit_transform(df["fuzzy_confidence"].values.reshape(-1, 1))
        return df

    def augment_ids(self, df, keep=None):
        xref_tab = self.h5_handle.root.XRef
        xrefs = pandas.DataFrame(
            xref_tab.read_where(
                "(EntryNr >= {}) & (EntryNr <= {}) & (XRefSource == {})".format(
                    self.genome_range[0],
                    self.genome_range[1],
                    xref_tab.get_enum("XRefSource")["SourceID"],
                )
            )
        )
        xrefs.drop(columns=["XRefSource", "Verification"], inplace=True)
        xrefs["XRefId"] = xrefs["XRefId"].str.decode("utf-8")

        if keep is None or keep == "first":
            xrefs_one_ref = xrefs.drop_duplicates(subset="EntryNr", keep="first")
        elif keep == "agg":
            xrefs_one_ref = xrefs.groupby("EntryNr")["XRefId"].agg(lambda x: "; ".join(x.values)).to_frame()
        else:
            raise ValueError('argument "keep" must be either "first" or "agg"')
        new_df = pandas.merge(
            pandas.merge(df, xrefs_one_ref, how="left", left_on="EntryNr1", right_on="EntryNr"),
            xrefs_one_ref,
            how="left",
            left_on="EntryNr2",
            right_on="EntryNr",
        )
        new_df.drop(columns=["EntryNr_x", "EntryNr_y"], inplace=True, errors="ignore")
        new_df.rename(columns={"XRefId_x": "SourceID1", "XRefId_y": "SourceID2"}, inplace=True)
        return new_df


class HomeologsConfidenceCalculatorFromTSV(HomeologsConfidenceCalculator):
    def __init__(self, infile):
        self.relations_df = pandas.read_csv(infile, sep="\t")
        expected_columns = [
            "EntryNr1",
            "EntryNr2",
            "SyntenyConservationLocal",
            "Distance",
        ]
        if len(set(expected_columns) - set(self.relations_df.columns.values)) > 0:
            raise KeyError(
                "provided inputfile does not have all expected columns. "
                "Expected columns are {}".format(expected_columns)
            )

    def augment_ids(self):
        raise NotImplementedError("cannot map to source ids from TSV input")


class MakeHomoeologDataFrame:
    """These methods get the homoeolog pairs and relevent information from the oma database, so you don't have to!"""

    def __init__(self, h5_handle, genome, **kwargs):
        self.h5_handle = h5_handle
        self.genome = genome
        self.kwargs = kwargs
        if isinstance(h5_handle, tables.File):
            self.h5_handle = h5_handle
        elif isinstance(h5_handle, (str, bytes)):
            self.h5_handle = tables.open_file(h5_handle, "r")
        else:
            raise TypeError("expected h5_handle to be either h5-file handle or a path to file")
        self.genome_row = next(self.h5_handle.root.Genome.where("UniProtSpeciesCode == genome"))
        self.genome_range = (
            int(self.genome_row["EntryOff"]) + 1,
            int(self.genome_row["EntryOff"] + self.genome_row["TotEntries"]),
        )

    def dedup_homoeolog_df(self, df):
        """Removes the double (inverse) homoeolog pairs from the table"""
        genome = self.genome
        if genome == "GOSHI":
            df = df[df["SubGenome1"] == "A"]
        if genome == "BRANA":
            df = df[df["SubGenome1"] == "A"]
        if genome == "WHEAT":
            df = df[
                (df["SubGenome1"] == "A") & (df["SubGenome2"] == "B")
                | (df["SubGenome1"] == "A") & (df["SubGenome2"] == "D")
                | (df["SubGenome1"] == "B") & (df["SubGenome2"] == "D")
            ]
        return df

    def get_entries_df(self):
        """Get the entries df belonging to a certain genome"""
        # Gets the first entrynr that corresponds to genome

        # Get all the entries of our genome and remove alternative splice variants
        entries_df = pandas.DataFrame(
            self.h5_handle.root.Protein.Entries.read_where(
                "(EntryNr >= {}) & (EntryNr <= {})".format(*self.genome_range)
            )
        )
        entries_df = entries_df[
            (entries_df["AltSpliceVariant"] == 0) | (entries_df["AltSpliceVariant"] == entries_df["EntryNr"])
        ]

        return entries_df

    def get_homoeolog_pairs_df(self, genome):
        # Create a dataframe containing only the homoeologs
        df = pandas.DataFrame(
            self.h5_handle.get_node("/PairwiseRelation/{}/within".format(self.genome)).read_where("RelType == 5")
        )
        return df

    def get_gene_info_for_homeologs(self, entries_df, homoeolog_pairs_df):
        # Merge with entries_df to get protein lengths, chromosomes, subgenomes for each entry
        df = pandas.merge(
            homoeolog_pairs_df,
            entries_df,
            how="left",
            left_on=["EntryNr1"],
            right_on=["EntryNr"],
        )
        df = pandas.merge(df, entries_df, how="left", left_on=["EntryNr2"], right_on=["EntryNr"])
        return df

    def decode_df(self, df):
        columns = df.columns
        for column in columns:
            try:
                df[column] = df[column].str.decode("utf-8")
            except:
                pass
        return df

    def add_genome_column(self, df):
        df.insert(loc=2, column="genome", value=genome)
        return df

    def _rename_and_remove_columns(self, df):
        df = df.rename(
            columns={
                "SyntenyConservationChromosome": "ConfidenceScore",
                "SyntenyConservationLocal": "SyntenyScore",
                "SeqBufferLength_x": "ProteinLen1",
                "SeqBufferLength_y": "ProteinLen2",
                "SubGenome_x": "SubGenome1",
                "SubGenome_y": "SubGenome2",
                "Chromosome_x": "Chr1",
                "Chromosome_y": "Chr2",
            }
        )
        # Remove irrelevant columns
        columns = [
            "EntryNr1",
            "EntryNr2",
            "Distance",
            "ConfidenceScore",
            "SyntenyScore",
            "TotalCopyNr",
            "ProteinLen1",
            "ProteinLen2",
            "Chr1",
            "Chr2",
            "SubGenome1",
            "SubGenome2",
        ]
        return df[columns]

    def add_dNdS_file_to_df(self, df, path_dNdS_file):
        # Only keep first column through dNdS column
        dNdS_dataframe = pandas.read_csv(path_dNdS_file, sep="\t").loc[:, :"dNdS"]
        # merge with homoeolog pair dataframe
        df = pandas.merge(
            df,
            dNdS_dataframe,
            how="left",
            left_on=["EntryNr1", "EntryNr2"],
            right_on=["entrynr1", "entrynr2"],
        )
        df.drop(["entrynr1", "entrynr2"], 1, inplace=True)
        return df

    def add_copynr(self, df):
        counts = collections.Counter(df["EntryNr1"])
        df["TotalCopyNr"] = df.apply(lambda x: counts[x["EntryNr1"]] + counts[x["EntryNr2"]], axis=1)
        return df

    def get_final_homoeolog_df(self):
        entries_df = self.get_entries_df()
        hom_df = self.get_homoeolog_pairs_df(self.genome)
        # remove ASVs
        df = hom_df[hom_df["EntryNr1"].isin(entries_df["EntryNr"]) & hom_df["EntryNr2"].isin(entries_df["EntryNr"])]
        # add number of duplications
        homObj = HomeologsConfidenceCalculator(self.h5_handle, self.genome)
        df = self.add_copynr(df)
        # Merge with entries_df to get protein lengths, chromosomes, subgenomes for each entry
        df = self.get_gene_info_for_homeologs(entries_df, df)
        df = self._rename_and_remove_columns(df)
        df = self.decode_df(df)
        df.insert(loc=2, column="Genome", value=self.genome)
        df = self.dedup_homoeolog_df(df)
        if "path_dNdS_file" in self.kwargs.keys():
            path_dNdS_file = self.kwargs["path_dNdS_file"]
            df = self.add_dNdS_file_to_df(df, path_dNdS_file)
            return df
        else:
            return df


if __name__ == "__main__":
    import argparse

    # Get arguments from command line
    parser = argparse.ArgumentParser(description="Computes homoeology confidence score using fuzzy logic")
    grp = parser.add_mutually_exclusive_group(required=True)
    grp.add_argument("--h5", help="name of hdf5 file, full path")
    grp.add_argument("--csv", help="tab-separated file with input data as alternative to hdf5 file")
    parser.add_argument(
        "--genome",
        help="5 letter code of polyploid genome to analyze. " "Must be specified if used with --h5 option.",
    )
    parser.add_argument(
        "--outfile",
        help="name where results will be stored (file name created to include parameters)",
        default="homoeolog_confidence.tsv",
    )

    args = parser.parse_args()
    logging.basicConfig(level=logging.INFO)

    if args.h5 is not None and args.genome is None:
        import sys

        sys.stderr.write("genomes argument required if using with an hdf5 file as input")
        sys.exit(1)

    if args.h5:
        import tables

        scorer = HomeologsConfidenceCalculator(tables.open_file(args.h5), args.genome)
    else:
        scorer = HomeologsConfidenceCalculatorFromTSV(args.csv)
    data = scorer.calculate_scores()
    try:
        data = scorer.augment_ids(data)
    except NotImplementedError:
        pass
    data.to_csv(args.outfile, sep="\t", header=True, index=True)



from __future__ import division, annotations

import collections
import re
from typing import List, Tuple

import numpy
import time
from .exceptions import UnknownSpecies, InvalidTaxonId, InvalidId
from .exceptions import Singleton as HOGSingleton


def format_sciname(sci, short=False):
    p = set(
        [
            sci.find(x)
            for x in [
                "(",
                "serogroup",
                "serotype",
                "serovar",
                "biotype",
                "subsp",
                "pv.",
                "bv.",
            ]
        ]
    )
    if sci.startswith("Escherichia coli"):
        p.add(sci.find("O"))
    p.discard(-1)
    p = min(p) if len(p) > 0 else len(sci)
    return {"species": sci[0:p], "strain": sci[p:]}


class LazyProperty(object):
    """Decorator to evaluate a property only on access.

    Compute the attribute value and caches it in the instance.
    Python Cookbook (Denis Otkidach) http://stackoverflow.com/users/168352/denis-otkidach
    This decorator allows you to create a property which can be computed once and
    accessed many times."""

    def __init__(self, method, name=None):
        # record the unbound-method and the name
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        # compute, cache and return the instance's attribute value
        result = self.method(inst)
        # setattr redefines the instance's attribute so this doesn't get called again
        setattr(inst, self.name, result)
        return result


class KeyWrapper(object):
    """
    Enables the use of functions, e.g. bisect, with a key function.
    """

    def __init__(self, it, key):
        self.it = it
        self.key = key

    def __getitem__(self, i):
        return self.key(self.it[i])

    def __len__(self):
        return len(self.it)


class Singleton(type):
    """A meta-class to enforce a Singleton, e.g. a class that can be
    instantiated only exactly once.

    Modified from Python Cookbook, 3rd Edition, p 357ff.

    :Example:

        class Foo(metaclass=Singleton):
            def __init__(self):
                pass  #This part is executed only once
    """

    def __init__(self, *args, **kwargs):
        self.__instance = None
        super(Singleton, self).__init__(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        if self.__instance is None:
            self.__instance = super(Singleton, self).__call__(*args, **kwargs)
        return self.__instance


class ProteinEntry(object):
    """Model for a protein object

    This class provides an easy to use interface for a given protein
    form the database.

    If instantiated with an entry_nr only, no data is loaded until a
    property or method is accessed. Properties that need to access
    additional data or loaded lazily and are cached in the object
    (but not kept after deletion of object)."""

    def __init__(self, db, e):
        self._stored_entry = e
        self._db = db

    @LazyProperty
    def _entry(self):
        return (
            self._db.entry_by_entry_nr(self._stored_entry)
            if isinstance(self._stored_entry, (int, numpy.integer))
            else self._stored_entry
        )

    @classmethod
    def from_entry_nr(cls, db, eNr):
        # e = db.entry_by_entry_nr(eNr)
        return cls(db, int(eNr))

    @property
    def entry_nr(self):
        if isinstance(self._stored_entry, (int, numpy.integer)):
            return int(self._stored_entry)
        return int(self._entry["EntryNr"])

    @property
    def locus_start(self):
        return int(self._entry["LocusStart"])

    @property
    def locus_end(self):
        return int(self._entry["LocusEnd"])

    @property
    def strand(self):
        return int(self._entry["LocusStrand"])

    @LazyProperty
    def exons(self):
        return ExonStructure.from_entry_nr(self._db, self.entry_nr)

    @property
    def nr_exons(self):
        return int(len(self.exons))

    @property
    def oma_group(self):
        return int(self._entry["OmaGroup"])

    @property
    def oma_hog(self):
        return self._entry["OmaHOG"].decode()

    @property
    def chromosome(self):
        return self._entry["Chromosome"].decode()

    @property
    def canonicalid(self):
        return self._entry["CanonicalId"].decode()

    @LazyProperty
    def xrefs(self):
        return self._db.id_mapper["Xref"].map_entry_nr(self._entry["EntryNr"])

    @property
    def sequence_md5(self):
        return self._entry["MD5ProteinHash"].decode()

    @LazyProperty
    def genome(self):
        g = self._db.id_mapper["OMA"].genome_of_entry_nr(self._entry["EntryNr"])
        return Genome(self._db, g)

    @LazyProperty
    def omaid(self):
        return self._db.id_mapper["OMA"].map_entry_nr(self._entry["EntryNr"])

    @property
    def oma_id(self):
        return self.omaid

    @LazyProperty
    def cdna(self):
        return self._db.get_cdna(self._entry).decode()

    @property
    def gc_content(self):
        cdna = self.cdna
        cnts = list(map(cdna.count, "GCAT"))
        try:
            return sum(cnts[0:2]) / sum(cnts)
        except ZeroDivisionError:
            return 0

    @LazyProperty
    def sequence(self):
        return self._db.get_sequence(self._entry).decode()

    @property
    def sequence_length(self):
        return int(self._entry["SeqBufferLength"]) - 1

    @LazyProperty
    def description(self):
        return self._db.get_description(self._entry).decode()

    @LazyProperty
    def ec_numbers(self):
        return self._db.get_ec_annotations(self.entry_nr)

    @property
    def subgenome(self):
        return self._entry["SubGenome"].decode()

    @LazyProperty
    def hog_family_nr(self):
        try:
            fam = self._db.hog_family(self._entry)
        except HOGSingleton:
            fam = 0
        return fam

    @LazyProperty
    def _subhog_parts(self) -> List[Tuple[int, str]]:
        root, *subhog_tokens = self.oma_hog.split(".")
        parsed_parts = []
        for part in subhog_tokens:
            if match := re.match(r"(\d+)([a-z]+)", part):
                num = int(match.group(1))
                char = match.group(2)
                parsed_parts.append((num, char))
            else:
                raise ValueError(f"Invalid part format in '{part}'")
        return parsed_parts

    def is_orthologous_to(self, other: ProteinEntry) -> bool | str:
        if self.entry_nr == other.entry_nr:
            return False
        elif self.hog_family_nr != other.hog_family_nr:
            return False
        else:
            common = [self.oma_hog.split(".")[0]]
            for (num1, char1), (num2, char2) in zip(self._subhog_parts, other._subhog_parts):
                if num1 != num2:
                    break
                elif char1 != char2:
                    return False
                common.append(f"{num1}{char1}")
            return ".".join(common)

    def is_paralogous_to(self, other: ProteinEntry) -> bool | str:
        if self.entry_nr == other.entry_nr:
            return False
        elif self.hog_family_nr != other.hog_family_nr:
            return False
        else:
            common = [self.oma_hog.split(".")[0]]
            for (num1, char1), (num2, char2) in zip(self._subhog_parts, other._subhog_parts):
                if num1 != num2:
                    return False
                elif char1 != char2:
                    return ".".join(common)
                common.append(f"{num1}{char1}")
            return False

    @property
    def is_main_isoform(self):
        return bool(self._entry["AltSpliceVariant"] == 0 or self._entry["AltSpliceVariant"] == self._entry["EntryNr"])

    @LazyProperty
    def alternative_isoforms(self):
        return [
            ProteinEntry(self._db, e)
            for e in self._db.get_splicing_variants(self._entry)
            if e["EntryNr"] != self.entry_nr
        ]

    def get_main_isoform(self):
        if self.is_main_isoform:
            return self
        else:
            return ProteinEntry(self._db, self._entry["AltSpliceVariant"])

    def __repr__(self):
        return "<{}({}, {})>".format(self.__class__.__name__, self.entry_nr, self.omaid)

    def __len__(self):
        return self.sequence_length


class Genome(object):
    """Model of a genome/proteome

    This model provides information about a genome. It is instantiated with
    row of the the /Genome table from the hdf5 file.
    Alternatively, it can now also be instantiated with a uniprot mnemonic
    species code or taxonomy id. Note that the validity of the parameter is
    only checked at first propery/method access.

    :param db: the underlying database object
    :type db: :class:`pyoma.browser.db.Database`
    :param genome: the genome data or identifier to be handled. See above for details.
    :type genome: (:class:`numpy.void`, int, str)"""

    def __init__(self, db, genome):
        self._stored_genome = genome
        self._db = db

    @LazyProperty
    def _genome(self):
        if isinstance(self._stored_genome, (numpy.void, numpy.ndarray)):
            return self._stored_genome
        else:
            return self._db.id_mapper["OMA"].identify_genome(self._stored_genome)

    @property
    def ncbi_taxon_id(self):
        """returns the ncbi taxonomy id of the genome"""
        return int(self._genome["NCBITaxonId"])

    @property
    def uniprot_species_code(self):
        """returns the uniprot mnemonic species code"""
        return self._genome["UniProtSpeciesCode"].decode()

    @property
    def sciname(self):
        """returns the scientific name of the genome"""
        return self._genome["SciName"].decode()

    @property
    def common_name(self):
        """returns the common name of the genome. If not set,
        the empty string is returned"""
        try:
            return self._genome["CommonName"].decode()
        except ValueError:
            return ""

    @property
    def synonym_name(self):
        """returns the synonym name"""
        return self._genome["SynName"].decode()

    @LazyProperty
    def species_and_strain_as_dict(self):
        return format_sciname(self.sciname)

    def species(self):
        return self.species_and_strain_as_dict["species"]

    def strain(self):
        return self.species_and_strain_as_dict["strain"]

    @property
    def url(self):
        try:
            return self._genome["Url"].decode()
        except ValueError:
            if self._db.db_schema_version < (3, 2):
                return ""
            raise

    @property
    def source(self):
        try:
            return self._genome["Source"].decode()
        except ValueError:
            if self._db.db_schema_version < (3, 2):
                return ""
            raise

    @property
    def release(self):
        return self._genome["Release"].decode()

    @property
    def last_modfied_timestamp(self):
        return self._genome["Date"]

    @property
    def last_modified(self):
        return self.modification_date("%Y-%b-%d")

    def modification_date(self, fmt):
        if self._db.db_schema_version >= (3, 2):
            return time.strftime(fmt, time.localtime(self.last_modfied_timestamp))
        else:
            return "n/a"

    @property
    def nr_entries(self):
        """returns the number of protein entries of the genome in the database"""
        return int(self._genome["TotEntries"])

    @LazyProperty
    def nr_genes(self):
        """returns the number of genes of the genome in the database"""
        try:
            return int(self._genome["TotGenes"])
        except ValueError:
            if self._db.db_schema_version < (3, 6):
                return self._db.count_main_isoforms(self.uniprot_species_code)
            raise

    @property
    def entry_nr_offset(self):
        return int(self._genome["EntryOff"])

    @LazyProperty
    def kingdom(self):
        # TODO: store directly in db
        return self._db.tax.get_parent_taxa(self._genome["NCBITaxonId"])[-1]["Name"].decode()

    @property
    def is_polyploid(self):
        """returns whether this is a polyploid genome"""
        return self._genome["IsPolyploid"]

    @LazyProperty
    def lineage(self):
        """returns the lineage of the genome."""
        return [lev["Name"].decode() for lev in self._db.tax.get_parent_taxa(self._genome["NCBITaxonId"])]

    @LazyProperty
    def chromosomes(self):
        chrs = collections.defaultdict(list)
        entry_tab = self._db.get_hdf5_handle().get_node("/Protein/Entries")
        for row in entry_tab.where(
            "(EntryNr > {}) & (EntryNr <= {})".format(self.entry_nr_offset, self.entry_nr_offset + self.nr_entries)
        ):
            chrs[row["Chromosome"].decode()].append(row["EntryNr"])
        return chrs

    def approx_chromosome_length(self, chromosome):
        """method to retrieve the approximate length of a given chromosome.
        The value corresponds to the end locus coordinate of the last gene
        on the chromosome.

        :param chromosome: the chromosome of interest
        :type chromosome: str, bytes
        :raises ValueError: if chromosome does not exist for this genome"""
        if isinstance(chromosome, bytes):
            chr = chromosome
        elif isinstance(chromosome, str):
            chr = chromosome.encode("utf-8")
        else:
            chr = "{}".format(chromosome).encode("utf-8")
        query = "(EntryNr > enr_off) & (EntryNr <= enr_max) & (Chromosome == chr)"
        condvars = {
            "enr_off": self.entry_nr_offset,
            "enr_max": self.entry_nr_offset + self.nr_entries,
            "chr": chr,
        }
        tab = self._db.get_hdf5_handle().get_node("/Protein/Entries")
        chr_len = int(max((row["LocusEnd"] for row in tab.where(query, condvars=condvars))))
        return chr_len

    def __repr__(self):
        return "<{}({}, {})>".format(self.__class__.__name__, self.uniprot_species_code, self.ncbi_taxon_id)

    def __len__(self):
        return self.nr_entries


class AncestralGenome(object):
    """Model of an ancestral genome/species

    This model provides information about an ancestral genome. It is instantiated with
    row of the the /Taxonomy table from the hdf5 file.
    Alternatively, it can now also be instantiated with a taxonomic id only.
    Note that the validity of the parameter is
    only checked at first propery/method access.

    :param db: the underlying database object
    :type db: :class:`pyoma.browser.db.Database`
    :param genome: the genome data or identifier to be handled. See above for details.
    :type genome: (:class:`numpy.void`, int, str)"""

    def __init__(self, db, genome):
        self.db = db
        self._stored_genome = genome

    @LazyProperty
    def _genome(self):
        if isinstance(self._stored_genome, (int, str, numpy.integer)):
            if (isinstance(self._stored_genome, (int, numpy.integer)) and self._stored_genome == 0) or (
                isinstance(self._stored_genome, str) and self._stored_genome.lower() == "luca"
            ):
                data = numpy.array([(0, -1, b"LUCA")], dtype=self.db.tax.tax_table.dtype)[0]
            else:
                try:
                    data = self.db.tax.get_taxnode_from_name_or_taxid(self._stored_genome)[0]
                except (InvalidTaxonId, KeyError) as e:
                    raise UnknownSpecies(
                        "No ancestral genome for {} exists in the database".format(self._stored_genome)
                    )
        else:
            data = self._stored_genome
        return data

    @property
    def ncbi_taxon_id(self):
        """the ncbi taxon id of this ancestral genome"""
        return int(self._genome["NCBITaxonId"])

    @property
    def scientific_name(self):
        """the scientific name of this ancestral genome"""
        return self._genome["Name"].decode()

    @property
    def sciname(self):
        return self.scientific_name

    @property
    def level(self):
        return self.scientific_name

    @LazyProperty
    def subtaxonomy(self):
        """the :class:`pyoma.browser.db.Taxonomy` instance of the sub taxonomy
        rooted at the node of the ancestral genome"""
        if self.ncbi_taxon_id == 0:
            return self.db.tax
        return self.db.tax.get_subtaxonomy_rooted_at(self.ncbi_taxon_id)

    @LazyProperty
    def extant_genomes(self):
        """the list of :class:`Genome` models instances representing all
        the extant genomes originating from this ancestral genome"""
        return [Genome(self.db, taxid) for taxid in self.subtaxonomy.get_taxid_of_extent_genomes()]

    @property
    def kingdom(self):
        return self.lineage[-1]

    @LazyProperty
    def lineage(self):
        """returns the lineage of the ancestral genome."""
        return [lev["Name"].decode() for lev in self.db.tax.get_parent_taxa(self.ncbi_taxon_id)]

    @LazyProperty
    def nr_genes(self):
        return self.db.count_hogs_at_level(self.scientific_name)


class OmaGroup(object):
    """OmaGroup object model

    The OmaGroup model can be instantiated with a group nr or a fingerprint.
    The (meta-)data for the group will be loaded lazily if properties are accessed."""

    def __init__(self, db, og):
        self._stored_group = og
        self._db = db

    @LazyProperty
    def _group(self):
        if isinstance(self._stored_group, dict) and "group_nr" in self._stored_group:
            return self._stored_group
        else:
            return self._db.oma_group_metadata(self._db.resolve_oma_group(self._stored_group))

    @property
    def group_nbr(self):
        """numeric representation of the OmaGroup"""
        if isinstance(self._stored_group, (int, numpy.integer)):
            if not 0 < self._stored_group <= self._db.get_nr_oma_groups():
                raise InvalidId("{} is an invalid oma group".format(self._stored_group))
            return int(self._stored_group)
        return int(self._group["group_nr"])

    @property
    def fingerprint(self):
        """fingerprint of the OmaGroup"""
        return self._group["fingerprint"]

    @property
    def keyword(self):
        """inferred keyword of the OmaGroup"""
        return self._group["keywords"]

    @LazyProperty
    def members(self):
        """returns a list of member proteins as :class:`ProteinEntry` objects"""
        return [ProteinEntry(self._db, pe) for pe in self._db.oma_group_members(self.group_nbr)]

    @property
    def nr_member_genes(self):
        """number of genes/proteins belonging to he group."""
        size = self._group["size"]
        if size < 0:
            # fallback in case it is not yet stored in db
            size = len(self.members)
        return size

    def __len__(self):
        return self.nr_member_genes

    def __repr__(self):
        return "<{}({}, {})>".format(self.__class__.__name__, self.group_nbr, self.fingerprint)


class HOG(object):
    """HOG object model

    This object stores information about a HOG at a certain level and provides
    convenince methods / properties to load various additional data related to
    the HOG from the underlying database.

    The model can be instantiated with a *hog* argument, that can be either a
    roothog family interger numer, or a valid Hog-ID (as string/byte). Optionally,
    the *level* argument can be used to select the sub-hog at the given taxonomic
    range. Up on access of any property / method of the
    created object, the model will lazily load the
    :class:`pyoma.browser.tablefmt.HogLevel` array for the specified hog / level
    arguments.

    :param db: the underlying database object
    :type db: :class:`pyoma.browser.db.Database`
    :param hog: the hog id or the HogLevel numpy array instance
    :type hog: int, str, bytes, numpy.void
    :param level: desired level of (sub-)HOG
    :type level: str, bytes, optional"""

    def __init__(self, db, hog, level=None):
        self._stored_hog = hog
        self._stored_level = level
        self._db = db

    @LazyProperty
    def _hog(self):
        if isinstance(self._stored_hog, (int, numpy.integer)):
            # load hog at root
            return self._db.get_hog(self._db.format_hogid(self._stored_hog), level=self._stored_level)
        elif isinstance(self._stored_hog, (str, bytes)):
            # load hog using ID
            return self._db.get_hog(self._stored_hog, self._stored_level)
        else:
            return self._stored_hog

    @property
    def fam(self):
        """roothog id as integer"""
        return int(self._hog["Fam"])

    @property
    def hog_id(self):
        """HogID of the HOG"""
        return self._hog["ID"].decode()

    @property
    def level(self):
        """taxonomic range of the HOG"""
        return self._hog["Level"].decode()

    @property
    def nr_member_genes(self):
        """number of extent genes belonging to the HOG"""
        return int(self._hog["NrMemberGenes"])

    @property
    def is_root(self):
        """whether this is the deepest taxonomic range for
        this (sub-)HOG."""
        return bool(self._hog["IsRoot"])

    @property
    def completeness_score(self):
        """returns the completness score of the HOG. It corresponds to the fraction of
        genomes that have at least one gene in this HOG."""
        return float(self._hog["CompletenessScore"])

    def __len__(self):
        return self.nr_member_genes

    def __repr__(self):
        return "<{}({}, {})>".format(self.__class__.__name__, self.hog_id, self.level)

    @LazyProperty
    def keyword(self):
        """returns the inferred keywords for the HOG. It is based on the descriptions
        and IDs of all the member genes"""
        return self._db.get_roothog_keywords(self.fam)

    @LazyProperty
    def members(self):
        """returns a list of :class:`pyoma.browser.models.ProteinEntry` instances with all
        the proteins / genes belonging to this HOG."""
        return [ProteinEntry(self._db, pe) for pe in self._db.member_of_hog_id(self.hog_id, level=self.level)]

    @LazyProperty
    def parent_hogs(self):
        """returns a list of the parent HOGs."""
        return self._db.get_parent_hogs(self.hog_id, self.level)


class PairwiseRelation(object):
    def __init__(self, db, relation, entry1=None, entry2=None):
        self._relation = relation
        self._db = db
        self._entry1 = entry1
        self._entry2 = entry2

    @property
    def distance(self):
        return float(self._relation["Distance"])

    @property
    def score(self):
        return float(self._relation["Score"])

    @property
    def alignment_overlap(self):
        return float(self._relation["AlignmentOverlap"])

    @property
    def synteny_conservation_local(self):
        return float(self._relation["SyntenyConservationLocal"])

    @property
    def confidence(self):
        return float(self._relation["Confidence"])

    @LazyProperty
    def rel_type(self):
        if not isinstance(self._relation["RelType"], str):
            type_map = self._db._get_pw_tab(self._relation["EntryNr1"], "VPairs").get_enum("RelType")
            return type_map(self._relation["RelType"])
        else:
            return self._relation["RelType"]

    @property
    def entry_1(self):
        if self._entry1 is None:
            self._entry1 = ProteinEntry(self._db, self._db.entry_by_entry_nr(self._relation["EntryNr1"]))
        return self._entry1

    @property
    def entry_2(self):
        if self._entry2 is None:
            self._entry2 = ProteinEntry(self._db, self._db.entry_by_entry_nr(self._relation["EntryNr2"]))
        return self._entry2


class GeneOntologyAnnotation(object):
    def __init__(self, db, anno):
        self.db = db
        self.anno = anno

    @LazyProperty
    def term(self):
        return self.db.gene_ontology.term_by_id(self.anno["TermNr"])

    @property
    def evidence(self):
        return self.anno["Evidence"].decode()

    @property
    def reference(self):
        return self.anno["Reference"].decode()

    @property
    def entry_nr(self):
        return int(self.anno["EntryNr"])

    @LazyProperty
    def object_id(self):
        return self.db.id_mapper["OMA"].map_entry_nr(self.entry_nr)

    @LazyProperty
    def ic(self):
        try:
            return self.db.freq_aware_gene_ontology.ic(self.term)
        except ValueError:
            return 0

    @LazyProperty
    def aspect(self):
        from .geneontology import GOAspect

        return GOAspect.to_string(self.term.aspect)


class ExonStructure(object):
    def __init__(self, db, exons):
        self._stored = exons
        self._db = db

    @LazyProperty
    def _exons(self):
        return self._db.get_exons(self._stored) if isinstance(self._stored, int) else self._stored

    @classmethod
    def from_entry_nr(cls, db, eNr):
        return cls(db, int(eNr))

    def _iter_exons(self):
        if len(self._exons) > 0 and self._exons["Strand"][0] < 0:
            self._exons[::-1].sort(order="Start")
        else:
            self._exons.sort(order="Start")
        for exon in self._exons:
            yield Exon(exon)

    def __len__(self):
        return len(self._exons)

    def __repr__(self):
        return "<{}(entry_nr={}, nr_exons={})>".format(self.__class__.__name__, self._exons[0]["EntryNr"], len(self))

    def __str__(self):
        exs = list(str(e) for e in self._iter_exons())
        if len(exs) > 1:
            return "join({})".format(", ".join(exs))
        elif len(exs) == 1:
            return exs[0]
        else:
            return "n/a"

    def as_list_of_dict(self):
        return [{"start": e.start, "end": e.end, "strand": e.strand} for e in self._iter_exons()]


class Exon(object):
    def __init__(self, exon):
        self.exon = exon

    @property
    def start(self):
        return int(self.exon["Start"])

    @property
    def end(self):
        return int(self.exon["End"])

    @property
    def strand(self):
        return "+" if self.exon["Strand"] > 0 else "-"

    def __str__(self):
        if self.exon["Strand"] < 0:
            template = "complement({}..{})"
        else:
            template = "{}..{}"
        return template.format(self.exon["Start"], self.exon["End"])



import pyham
import xml.etree.ElementTree as ET
import logging

logger = logging.getLogger(__name__)


def get_ham_treemap_from_row(row, tree):
    """get the mapping of a pyham tree profile for a fiven root hog"""
    fam, orthoxml = row
    orthoxml = switch_name_ncbi_id(orthoxml)
    try:
        ham_obj = pyham.Ham(
            tree,
            orthoxml,
            type_hog_file="orthoxml",
            use_internal_name=True,
            orthoXML_as_string=True,
        )
        tp = ham_obj.create_tree_profile(hog=ham_obj.get_list_top_level_hogs()[0])
        return tp.treemap
    except (TypeError, AttributeError) as err:
        logger.exception("Error in pyham treeprofile occured")
        return None


def switch_name_ncbi_id(orthoxml):
    """swap ncbi taxid for species name to avoid ambiguity"""
    root = ET.fromstring(orthoxml)
    for node in root:
        if "species" in node.tag:
            node.attrib["name"] = node.attrib["NCBITaxId"]
    orthoxml = ET.tostring(root, encoding="unicode", method="xml")
    return orthoxml



from __future__ import unicode_literals, division, absolute_import

import sys
import time
import unittest
from unittest.mock import MagicMock
from builtins import str

import numpy
from future.utils import with_metaclass

import pyoma.browser.exceptions
from pyoma.browser import models, db, tablefmt
from .test_db import find_path_to_test_db


class TestDbBase(unittest.TestCase):
    db = None

    @classmethod
    def setUpClass(cls):
        fn = find_path_to_test_db("TestDb.h5")
        cls.db = db.Database(fn)

    @classmethod
    def tearDownClass(cls):
        cls.db.close()


class ProteinEntryTests(TestDbBase):
    def test_init_from_enr(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        self.assertEqual(protein_entry.entry_nr, 12)
        try:
            protein_entry.canonicalid
        except AttributeError:
            self.assertTrue(False, "entry not properly loaded")

    def test_lazy_eval_of_properties(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        self.assertNotIn("sequence", protein_entry.__dict__)
        seq = protein_entry.sequence
        self.assertIn("sequence", protein_entry.__dict__)

    def test_repr_of_ProteinEntry(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        if sys.version_info[0] >= 3 and sys.version_info[1] > 2:
            self.assertRegex("{!r}".format(protein_entry), r"<ProteinEntry\(12,.*\)")
        else:
            self.assertRegexpMatches("{!r}".format(protein_entry), r"<ProteinEntry\(12,.*\)")

    def test_len_of_entry(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        self.assertEqual(len(protein_entry), len(protein_entry.sequence))

    def test_ec_content(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        self.assertLessEqual(protein_entry.gc_content, 1)
        self.assertGreaterEqual(protein_entry.gc_content, 0)

    def test_example_ec_content(self):
        protein_entry = models.ProteinEntry.from_entry_nr(self.db, 12)
        protein_entry.cdna = "GCGAATAT"
        self.assertAlmostEqual(protein_entry.gc_content, 3.0 / 8.0)


class HOGModelTest(TestDbBase):
    def test_instant_with_fam(self):
        hog = models.HOG(self.db, 2)
        self.assertEqual(self.db.format_hogid(2), hog.hog_id)
        self.assertTrue(hog.is_root)

    def test_instance_from_hogid(self):
        hog = models.HOG(self.db, self.db.format_hogid(1))
        self.assertEqual(1, hog.fam)
        self.assertTrue(hog.is_root)

    def test_invalid_level(self):
        with self.assertRaises(ValueError):
            hog = models.HOG(self.db, 2, "Metazoa")
            hog.hog_id

    def test_with_valid_level(self):
        hog = models.HOG(self.db, 2, "Fungi")
        self.assertEqual("Fungi", hog.level)
        self.assertFalse(hog.is_root)

    def test_members_and_nr_members(self):
        hog = models.HOG(self.db, "HOG:0000002.1a")
        self.assertEqual(2, hog.fam)
        self.assertEqual(hog.nr_member_genes, len(hog.members))

    def test_keyword_of_hog(self):
        hog = models.HOG(self.db, 2)
        self.assertEqual("", hog.keyword)


class TestProteinEntryOrthologyParalogy(unittest.TestCase):
    def setUp(self):
        # Mock database and entry data
        self.mock_db = MagicMock()
        # Minimal entry dicts with required fields
        self.entry1 = {"EntryNr": 1, "OmaHOG": b"1.15a.2bz"}
        self.entry2 = {"EntryNr": 2, "OmaHOG": b"1.15a.2b"}
        self.entry3 = {
            "EntryNr": 3,
            "OmaHOG": b"1.15b.2b",
        }
        self.entry4 = {
            "EntryNr": 4,
            "OmaHOG": b"5.1b",
        }
        self.entry5 = {
            "EntryNr": 5,
            "OmaHOG": b"5",
        }
        self.entry6 = {
            "EntryNr": 6,
            "OmaHOG": b"",
        }
        self.entry7 = {
            "EntryNr": 7,
            "OmaHOG": b"",
        }

        def hog_family_sideffect(entry):
            f = entry["OmaHOG"].split(b".")[0]
            if len(f) == 0:
                raise pyoma.browser.exceptions.Singleton(entry)
            return int(f)

            # Patch hog_family method

        self.mock_db.hog_family.side_effect = hog_family_sideffect

    def make_entry(self, entry_dict):
        # Patch id_mapper and other required db methods if needed
        return models.ProteinEntry(self.mock_db, entry_dict)

    def test_not_orthologous_if_same(self):
        p1 = self.make_entry(self.entry1)
        self.assertFalse(p1.is_orthologous_to(p1))

    def test_not_paralogous_if_same(self):
        p1 = self.make_entry(self.entry1)
        self.assertFalse(p1.is_paralogous_to(p1))

    def test_not_orthologous_but_paralogous_if_different_subhogid(self):
        p1 = self.make_entry(self.entry1)
        p2 = self.make_entry(self.entry2)
        self.assertFalse(p1.is_orthologous_to(p2))
        self.assertFalse(p2.is_orthologous_to(p1))
        self.assertEqual("1.15a", p1.is_paralogous_to(p2))
        self.assertEqual("1.15a", p2.is_paralogous_to(p1))

    def test_not_orthologous_if_different_family(self):
        p1 = self.make_entry(self.entry1)
        p4 = self.make_entry(self.entry4)
        self.assertFalse(p1.is_orthologous_to(p4))

    def test_subfam_orthologous_to_superfam(self):
        p4 = self.make_entry(self.entry4)
        p5 = self.make_entry(self.entry5)
        self.assertEqual("5", p4.is_orthologous_to(p5))
        self.assertEqual("5", p5.is_orthologous_to(p4))

    def test_subfam_are_not_paralogous_to_superfam(self):
        p4 = self.make_entry(self.entry4)
        p5 = self.make_entry(self.entry5)
        self.assertFalse(p4.is_paralogous_to(p5))

    def test_paralog_not_ortholog_on_subhog(self):
        p2 = self.make_entry(self.entry2)
        p3 = self.make_entry(self.entry3)
        self.assertFalse(p2.is_orthologous_to(p3))
        self.assertFalse(p3.is_orthologous_to(p2))
        self.assertEqual("1", p2.is_paralogous_to(p3))
        self.assertEqual("1", p3.is_paralogous_to(p2))

    def test_neither_paralog_nor_ortholog_on_subhog_with_empty_hogid(self):
        p6 = self.make_entry(self.entry6)
        p3 = self.make_entry(self.entry3)
        self.assertFalse(p6.is_orthologous_to(p3))
        self.assertFalse(p3.is_paralogous_to(p6))

    def test_orthologous_to_paralogous_on_singletons(self):
        p6 = self.make_entry(self.entry6)
        p7 = self.make_entry(self.entry7)
        self.assertFalse(p6.is_orthologous_to(p7))
        self.assertFalse(p7.is_paralogous_to(p6))


class OmaGroupModelTest(TestDbBase):
    def test_instant_with_grpnr(self):
        grp = models.OmaGroup(self.db, 1384)
        self.assertEqual(grp.group_nbr, 1384)

    def test_raises_invalid_id_for_invalid_group_nrs(self):
        for invalid_grp_nr in (-1, 0, 5421):
            with self.assertRaises(pyoma.browser.exceptions.InvalidId):
                grp = models.OmaGroup(self.db, invalid_grp_nr)
                grp.group_nbr

    def test_instant_with_fingerprint(self):
        grp = models.OmaGroup(self.db, "SDNEIRR")
        self.assertEqual("SDNEIRR", grp.fingerprint)

    def test_grp_size(self):
        for grpnr in (521, 4125, 532, 12):
            grp = models.OmaGroup(self.db, grpnr)
            self.assertEqual(len(grp), len(grp.members))

    def test_keyword(self):
        grp = models.OmaGroup(self.db, 521)
        self.assertEqual("methylenetetrahydrofolate reductase", grp.keyword)


class GenomeModelTest(TestDbBase):
    def test_count_genes(self):
        g = self.db.id_mapper["OMA"].identify_genome("YEAST")
        genome = models.Genome(self.db, g)
        self.assertGreaterEqual(genome.nr_genes, genome.nr_entries)

    def test_nr_genes_adds_up(self):
        genomes = self.db.id_mapper["OMA"].genome_table
        tot_genes = sum(models.Genome(self.db, g).nr_genes for g in genomes)
        self.assertEqual(tot_genes, self.db.count_main_isoforms())

    def test_chromosome_length(self):
        g = self.db.id_mapper["OMA"].identify_genome("YEAST")
        genome = models.Genome(self.db, g)
        chr_len = genome.approx_chromosome_length("I")
        self.assertGreaterEqual(chr_len, 220000)
        self.assertLessEqual(chr_len, 230218)  # actual len according to Ensembl

    def test_instantiate_from_species_code(self):
        genome = models.Genome(self.db, "YEAST")
        self.assertEqual(genome.uniprot_species_code, "YEAST")

    def test_raises_for_invalid_species(self):
        with self.assertRaises(pyoma.browser.exceptions.UnknownSpecies):
            genome = models.Genome(self.db, "HUMAN")
            genome.uniprot_species_code

    def test_instantitate_from_taxid(self):
        genome = models.Genome(self.db, 559292)
        self.assertEqual(genome.uniprot_species_code, "YEAST")


class AncestralGenomeModelTests(TestDbBase):
    def test_ncbi_taxon_id_agrees(self):
        query = 4890
        ag = models.AncestralGenome(self.db, query)
        self.assertEqual(query, ag.ncbi_taxon_id)

    def test_retrieve_extant_genomes(self):
        query = 4890
        ag = models.AncestralGenome(self.db, query)
        self.assertIn("YEAST", (g.uniprot_species_code for g in ag.extant_genomes))

    def test_lineage(self):
        query = 4890
        ag = models.AncestralGenome(self.db, query)
        self.assertEqual(ag.lineage[0], ag.sciname)
        self.assertEqual("Eukaryota", ag.kingdom)

    def test_raises_UnknownSpecies_for_invalid_taxid(self):
        with self.assertRaises(pyoma.browser.exceptions.UnknownSpecies):
            tax66 = models.AncestralGenome(self.db, 66).ncbi_taxon_id

    def test_with_luca(self):
        query = "LUCA"
        ag = models.AncestralGenome(self.db, query)
        self.assertEqual(0, ag.ncbi_taxon_id)

    def test_extant_genomes_of_luca(self):
        self.assertEqual(4, len(models.AncestralGenome(self.db, 0).extant_genomes))

    def test_extant_genome_of_extant_genome_taxid(self):
        query = 559292
        ag = models.AncestralGenome(self.db, query)
        self.assertIn(query, [g.ncbi_taxon_id for g in ag.extant_genomes])

    def test_nr_genes(self):
        query_level = "Eukaryota"
        ag = models.AncestralGenome(self.db, query_level)
        self.assertLess(100, ag.nr_genes)


class PairwiseRelationModelTest(TestDbBase):
    def test_relations(self):
        tab_data = self.db.get_vpairs_between_species_pair("YEAST", "PLAF7")
        rels = [models.PairwiseRelation(self.db, d) for d in tab_data]
        self.assertEqual(tab_data[0]["EntryNr1"], rels[0].entry_1.entry_nr)
        self.assertEqual(tab_data[0]["Distance"], rels[0].distance)

    def test_faster_when_passing_entries(self):
        tab_data = self.db.get_vpairs_between_species_pair("YEAST", "PLAF7")
        entries = {e["EntryNr"]: models.ProteinEntry(self.db, e) for e in self.db.main_isoforms("YEAST")}

        t0 = time.time()
        rels = [models.PairwiseRelation(self.db, d) for d in tab_data]
        entry1_nrs_1 = [r.entry_1.entry_nr for r in rels]
        t1 = time.time()
        rels2 = [models.PairwiseRelation(self.db, d, entry1=entries[d["EntryNr1"]]) for d in tab_data]
        entry1_nrs_2 = [r.entry_1.entry_nr for r in rels2]
        t2 = time.time()
        self.assertEqual(entry1_nrs_1, entry1_nrs_2)
        print(t2 - t1, "vs", t1 - t0, "seconds elapsed")
        self.assertLessEqual(t2 - t1, t1 - t0)


class ExonStructureTest(unittest.TestCase):
    def get_exons(self):
        loc_dtype = tablefmt.tables.dtype_from_descr(tablefmt.LocusTable)
        return numpy.array([(1, 500, 510, 1), (1, 600, 610, 1)], dtype=loc_dtype)

    def test_exon_struct_len(self):
        ex = models.ExonStructure(None, self.get_exons())
        self.assertEqual(2, len(ex))

    def test_str_repr(self):
        ex = models.ExonStructure(None, self.get_exons())
        self.assertEqual("join(500..510, 600..610)", str(ex))

    def test_json_repr(self):
        ex = models.ExonStructure(None, self.get_exons())
        self.assertEqual(
            [
                {"start": 500, "end": 510, "strand": "+"},
                {"start": 600, "end": 610, "strand": "+"},
            ],
            ex.as_list_of_dict(),
        )

    def test_str_repr_if_reverse_complement(self):
        ex_dat = self.get_exons()
        ex_dat["Strand"] = -1
        ex = models.ExonStructure(None, ex_dat)
        self.assertEqual("join(complement(600..610), complement(500..510))", str(ex))


class GeneOntologyAnnotationTests(unittest.TestCase):
    db = None

    @classmethod
    def setUpClass(cls):
        fn = find_path_to_test_db("TestDb.h5")
        cls.db = db.Database(fn)

    @classmethod
    def tearDownClass(cls):
        cls.db.get_hdf5_handle().close()

    def test_from_annotation_table(self):
        annos = self.db.get_gene_ontology_annotations(12)
        goa = models.GeneOntologyAnnotation(self.db, annos[0])
        self.assertEqual(12, goa.entry_nr)
        self.assertIn(
            goa.aspect,
            ["molecular_function", "biological_process", "cellular_component"],
        )
        self.assertEqual("YEAST00012", goa.object_id)


class SingletonTests(unittest.TestCase):
    def test_singleton(self):
        class Foo(with_metaclass(models.Singleton, object)):
            def __init__(self):
                pass

        a = Foo()
        b = Foo()
        self.assertEqual(a, b)
